@@ -3,11 +3,20 @@
 
 
 
+
 #undef DEBUG 
 #undef SHOW_RETRIES
 #undef SHOW_RTT_REMOVAL
 
 #define NUM_RETRIES 2
+#define SCAN_FATAL_ERR (void*)(-1)
+
+#ifndef ipci
+#define ipci()
+#define ipcd()
+#define ipcc() ( 0 == 1 )
+#define cct(x,y) ( 1 == 1 )
+#endif
 
 /*----------------------------------------------------------------------------*/
 struct pseudohdr {
@@ -20,6 +29,19 @@ struct pseudohdr {
 };
 
 
+#ifndef IPV6_SUPPORT
+#define sockaddr_in6 sockaddr
+#define in_addr6     in_addr
+#ifndef AF_INET6
+#define AF_INET6     AF_INET
+#endif
+#endif
+
+
+static int is_ipv6 = 0;
+
+
+
 static int 
 in_cksum(p, n)
 	u_short        *p;
@@ -68,12 +90,25 @@ struct timeval
 timeval(unsigned long val)
 {
 	struct timeval  ret;
+	unsigned int h, l;
 
 	val = ntohl(val);
 
-	ret.tv_sec = ((val & 0xF0000000) >> 28);
-	ret.tv_usec = ((val & 0x0FFFFFFF) << 4);
+	h = ( val & 0xF0000000 ) >> 28;
+	l = ( val & 0x0FFFFFFF)  << 4;
+     
+	ret.tv_sec = h;
+	ret.tv_usec = l;
+	while ( ret.tv_usec >= 1000000 ) 
+	 {
+	  ret.tv_usec -= 1000000;
+	  ret.tv_sec ++;
+	 }
 
+	if ( ret.tv_sec > 2 ) {
+	 ret.tv_sec = 2;
+	 ret.tv_usec = 0;
+	}
 	return ret;
 }
 
@@ -89,33 +124,32 @@ compute_rtt(unsigned long then)
 	unsigned long   now = maketime();
 	unsigned long   res;
 	unsigned long   a, b;
-#if DEBUG > 4
-	printf("now = 0x%.8x, then = 0x%.8x\n", ntohl(now), ntohl(then));
-#endif
+
 	a = (unsigned long) ntohl(now);
 	b = (unsigned long) ntohl(then);
 
+
 	if (b > a) {
 		return 0;
 	}
 	res = a - b;
+	if ( res >= (1 << 28) ) 
+		res = 1 << 28;
 
-        return htonl(res + 5000);
+        return htonl(res);
 }
 
 
 int 
-packetdead(unsigned long then, unsigned long rtt)
+packetdead(unsigned long then, unsigned long rtt, unsigned long now)
 {
-	unsigned long   now = maketime();
-
 	then = ntohl(then);
 	now = ntohl(now);
 	rtt = ntohl(rtt);
 
 
 
-	if ((now - then) > rtt) {
+	if ((now - then) >= 2 << 28 ) {
 		return 1;
 	} else {
 		return 0;
@@ -125,20 +159,27 @@ packetdead(unsigned long then, unsigned long rtt)
 int 
 rawsocket()
 {
-	int             soc = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
+	int             soc = -1;
 	int             opt = 1;
 
-	if (soc < 0) {
+
+	if ( is_ipv6 == 0 )
+ 	{
+	 soc = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
+	 if (soc < 0) {
 		perror("socket ");
 		return -1;
-	}
+	 }
 #ifdef IP_HDRINCL
-	if (setsockopt(soc, IPPROTO_IP, IP_HDRINCL, (char *) &opt, sizeof(opt)) < 0) {
+	 if (setsockopt(soc, IPPROTO_IP, IP_HDRINCL, (char *) &opt, sizeof(opt)) < 0) {
 		perror("setsockopt ");
 		close(soc);
 		return -1;
-	}
+	 }
 #endif
+	}
+	else
+  		soc = socket(AF_INET6, SOCK_RAW, IPPROTO_TCP);
 
 	return soc;
 }
@@ -157,6 +198,32 @@ openbpf(struct in_addr dst, struct in_addr * src, int magic)
 	return bpf;
 }
 
+#ifdef IPV6_SUPPORT
+int 
+openbpf6(struct sockaddr_in6 * dst, int magic)
+{
+	char           *iface;
+	char            filter[255];
+	char 		buf[512];
+	int             bpf;
+	struct 	in6_addr src;
+
+	
+	iface = routethrough6(dst, &src);
+	inet_ntop(AF_INET6, &dst->sin6_addr, buf, sizeof(buf));
+	snprintf(filter, sizeof(filter), "ip6 and tcp and src host %s and dst port %d", buf, magic);
+	bpf = bpf_open_live(iface, filter);
+	return bpf;
+}
+#else
+int 
+openbpf6(struct sockaddr_in6 * dst, int magic)
+{
+ return -1;
+}
+#endif
+
+
 /*----------------------------------------------------------------------------*/
 
 struct list {
@@ -170,6 +237,7 @@ struct list {
 struct list    *
 get_packet(struct list * l, unsigned short dport)
 {
+	if ( l == SCAN_FATAL_ERR ) return NULL;
 	while (l != NULL) {
 		if (l->dport == dport)
 			return l;
@@ -185,6 +253,7 @@ add_packet(struct list * l, unsigned short dport, unsigned long ack)
 {
 	struct list    *ret;
 
+	if ( l == SCAN_FATAL_ERR ) return NULL;
 	ret = get_packet(l, dport);
 	if (ret != NULL) {
 #ifdef SHOW_RETRIES
@@ -216,6 +285,7 @@ rm_packet(struct list * l, unsigned short dport)
 	struct list    *ret = l;
 	struct list    *p = get_packet(l, dport);
 
+	if ( l == SCAN_FATAL_ERR ) return NULL;
 	if (p == NULL) {
 #if DEBUG > 1
 		fprintf(stderr, "Odd - no entry for %d - RTT too low ?!\n", dport);
@@ -239,21 +309,25 @@ rm_dead_packets(struct list * l, unsigned long rtt, int *retry)
 {
 	struct list    *ret = l;
 	struct list    *p = l;
+	unsigned long now = maketime();
 
 
+	if ( l == SCAN_FATAL_ERR ) return NULL;
 	*retry = 0;
 	while (p != NULL) {
 		struct list    *next = p->next;
-		if (packetdead(p->when, rtt)) {
+		if (packetdead(p->when, rtt, now)) {
 			if (p->retries < NUM_RETRIES) {
 #ifdef SHOW_RETRIES
 				printf("Will retry port %d\n", p->dport);
 #endif
 				*retry = p->dport;
+				return ret;
 			} else {
 #ifdef SHOW_RTT_REMOVAL
 				printf("Removing port %d (RTT elapsed)\n", p->dport);
 #endif
+				ipcd();
 				if (p->next != NULL)
 					p->next->prev = p->prev;
 
@@ -278,14 +352,29 @@ rm_dead_packets(struct list * l, unsigned long rtt, int *retry)
 struct tcphdr * extracttcp(char * pkt, int len)
 {
  struct ip * ip;
-	struct tcphdr  *tcp;
-	unsigned long   ret;
+ struct tcphdr  *tcp;
+ unsigned long   ret;
+ 
+
+ if ( is_ipv6 == 0 && ( pkt[0] & 0xF0 ) != 0x40 ) return NULL;
+ else if ( is_ipv6 != 0 && ( pkt[0] & 0xF0 ) != 0x60 ) return NULL;
+
+
+ if ( is_ipv6 == 0 )
+ {
+  ip = (struct ip*)pkt;
+  if(ip->ip_hl * 4 + sizeof(struct tcphdr) > len)
+   return NULL;
+  tcp = (struct tcphdr*)(pkt + ip->ip_hl * 4);
+ }
+ else
+ {
+  if ( pkt[6] != IPPROTO_TCP || len < 60 ) return NULL; 
+  
+  tcp = (struct tcphdr*)( pkt + 40 );
+ }
  
- ip = (struct ip*)pkt;
- if(ip->ip_hl * 4 + sizeof(struct tcphdr) > len)
-  return NULL;
   
- tcp = (struct tcphdr*)(pkt + ip->ip_hl * 4);
  return tcp;
 }
 
@@ -302,11 +391,24 @@ extractack(char *pkt, int len)
 	return ret;
 }
 
+unsigned long 
+extractseq(char *pkt, int len)
+{
+ unsigned long   ret;
+ struct tcphdr * tcp = extracttcp(pkt, len);
+ 
+ if( tcp == NULL )
+  return -1;
+ 
+ ret = tcp->th_seq;
+	return ret;
+}
+
 
 unsigned short 
 extractsport(char *pkt, int len)
 {
-	struct tcphdr  *tcp = extracttcp(pkt, len);
+ struct tcphdr  *tcp = extracttcp(pkt, len);
  unsigned long   ret;
  if(tcp == NULL)return 0;
  
@@ -321,11 +423,12 @@ issynack(char *pkt, int len)
  
  if(tcp == NULL)return 0;
 
-	return tcp->th_flags == (TH_SYN | TH_ACK);
+	return (tcp->th_flags & (TH_SYN|TH_ACK)) == (TH_SYN|TH_ACK) &&
+               (tcp->th_flags  & (TH_RST|TH_FIN)) == 0;
 }
 
-char           *
-mktcp(struct in_addr src, int sport, struct in_addr dst, int dport, unsigned long th_ack)
+char *
+mktcp(struct in_addr src, int sport, struct in_addr dst, int dport, unsigned long th_seq, unsigned long th_ack, unsigned char flag, int * len)
 {
 	static char     pkt[sizeof(struct ip) + sizeof(struct tcphdr)];
 	struct ip      *ip;
@@ -350,11 +453,11 @@ mktcp(struct in_addr src, int sport, struct in_addr dst, int dport, unsigned lon
 	tcp = (struct tcphdr *) (&(pkt[sizeof(struct ip)]));
 	tcp->th_sport = htons(sport);
 	tcp->th_dport = htons(dport);
-	tcp->th_seq = th_ack;
-	tcp->th_ack = 0;
+	tcp->th_seq = th_seq;
+	tcp->th_ack = th_ack;
 	tcp->th_x2 = 0;
 	tcp->th_off = 5;
-	tcp->th_flags = TH_SYN;
+	tcp->th_flags = flag;
 	tcp->th_win = 4096;
 	tcp->th_sum = 0;
 	tcp->th_urp = 0;
@@ -367,13 +470,56 @@ mktcp(struct in_addr src, int sport, struct in_addr dst, int dport, unsigned lon
 	bcopy((char *) tcp, (char *) &pseudohdr.tcpheader, sizeof(struct tcphdr));
 	bcopy(&pseudohdr, tcpsumdata, sizeof(struct pseudohdr));
 	tcp->th_sum = in_cksum((unsigned short *) tcpsumdata, 12 + sizeof(struct tcphdr));
-
+	*len = 40;
 	return pkt;
 }
 /*--------------------------------------------------------------------*/
+#ifdef IPV6_SUPPORT
+char * mktcp6(struct sockaddr_in6 * dst, int sport, int dport, unsigned long th_seq, unsigned long th_ack, unsigned char flag, int * len)
+{
+	static char     pkt[sizeof(struct tcphdr)];
+	struct tcphdr  *tcp;
+	struct in6_addr source;
+	unsigned char buf[60];
 
-int 
-find_rtt(struct in_addr dst, unsigned long *rtt)
+
+        bzero(pkt, sizeof(pkt));
+	bzero(buf, sizeof(buf));
+
+	tcp = (struct tcphdr *) (pkt);
+	tcp->th_sport = htons(sport);
+	tcp->th_dport = htons(dport);
+	tcp->th_seq = th_seq;
+	tcp->th_ack = th_ack;
+	tcp->th_x2 = 0;
+	tcp->th_off = 5;
+	tcp->th_flags = flag;
+	tcp->th_win = 4096;
+	tcp->th_sum = 0;
+	tcp->th_urp = 0;
+
+ 	routethrough6(dst, &source);
+	memcpy(&(buf[0]), &source, sizeof(source));
+	memcpy(&(buf[16]), &dst->sin6_addr, sizeof(dst->sin6_addr));
+	buf[34] = sizeof(struct tcphdr) / 256;
+	buf[35] = sizeof(struct tcphdr) % 256;
+	buf[39] = IPPROTO_TCP;
+	memcpy(&(buf[40]), tcp, 20);
+	tcp->th_sum = in_cksum(buf, 60);
+	*len = 20;
+
+	return pkt;
+}
+#else
+char * mktcp6(struct sockaddr_in6 * dst, int sport, int dport, unsigned long th_seq, unsigned long th_ack, unsigned char flag, int * len)
+{
+ return NULL;
+}
+#endif
+
+
+/*--------------------------------------------------------------------*/
+int find_rtt(struct in_addr dst, struct sockaddr_in6 * sin6, unsigned long *rtt)
 {
 	int             soc;
 	unsigned short  ports[] = {21, 22, 34, 25, 53, 79, 80, 110, 113, 135, 139, 143, 264, 389, 443, 993, 1454, 1723, 3389, 8080, 0};
@@ -391,13 +537,18 @@ find_rtt(struct in_addr dst, unsigned long *rtt)
 	unsigned long   max, max_max;
 	int             err = 0;
 	int             noresend = 0;
+	struct sockaddr * so;
+	int		  so_sz;
 
 
 	soc = rawsocket();
 	if (soc < 0)
 		return -1;
 
-	bpf = openbpf(dst, &src, magic);
+	if ( is_ipv6 != 0 )
+	   bpf = openbpf6(sin6, magic);
+	else
+	  bpf = openbpf(dst, &src, magic);
 
 	if (bpf < 0) {
 		close(soc);
@@ -405,25 +556,36 @@ find_rtt(struct in_addr dst, unsigned long *rtt)
 	}
 	skip = get_datalink_size(bpf_datalink(bpf));
 	bzero(&soca, sizeof(soca));
-	soca.sin_family = AF_INET;
-	soca.sin_addr = dst;
+	if ( is_ipv6 == 0 )
+	{
+	 soca.sin_family = AF_INET;
+	 soca.sin_addr = dst;
+	 so = (struct sockaddr*)&soca;
+	 so_sz = sizeof(soca);
+	}
+	else {
+		so = (struct sockaddr*)sin6;
+		so_sz = sizeof(*sin6);
+	}
 
 	for (i = 0; ports[i] != 0; i++) {
 		char           *res;
 		unsigned long   ack = maketime();
-		char           *pkt = mktcp(src, magic, dst, ports[i], ack);
+		int		plen = 0;
+		char           *pkt = (is_ipv6 == 0) ? mktcp(src, magic, dst, ports[i], ack, 0, TH_SYN, &plen) : mktcp6(sin6, magic, ports[i], ack, 0, TH_SYN, &plen);
 		int             e;
-		struct timeval  tv = {2, 0};
+		struct timeval  tv = {1, 0};
 		int             err = 0;
 		unsigned short  p = ports[i];
 
 #if DEBUG > 1
 		printf("send to port %d\n", ports[i]);
 #endif
-
-		e = sendto(soc, pkt, sizeof(struct ip) + sizeof(struct tcphdr), 0, (struct sockaddr *) & soca, sizeof(soca));
+		
+		errno = 0;
+		e = sendto(soc, pkt, plen, 0, so, so_sz);
 		if (e < 0) {
-			perror("sendto ");
+			perror("synscan->find_rtt->sendto ");
 			close(soc);
 			bpf_close(bpf);
 			return -1;
@@ -445,7 +607,7 @@ find_rtt(struct in_addr dst, unsigned long *rtt)
 #endif
 		bpf_close(bpf);
 		close(soc);
-		*rtt = htonl(2 << 28);	/* Two seconds */
+		*rtt = htonl(1 << 28);	/* One second */
 		return 0;
 	}
 	max = max_max = 0;
@@ -454,14 +616,22 @@ find_rtt(struct in_addr dst, unsigned long *rtt)
 		char           *res;
 		unsigned long   ack = maketime();
 		char           *pkt;
+		int		plen;
 		int             e;
-		struct timeval  tv = {2, 0};
+		struct timeval  tv = {1, 0};
 
 
-		pkt = mktcp(src, magic, dst, use[n % num], ack);
-		e = sendto(soc, pkt, sizeof(struct ip) + sizeof(struct tcphdr), 0, (struct sockaddr *) & soca, sizeof(soca));
+		pkt = (is_ipv6 == 0) ? mktcp(src, magic, dst, use[n % num], ack, 0, TH_SYN, &plen) :
+				       mktcp6(sin6, magic, use[n%num], ack, 0, TH_SYN, &plen);
+
+		e = sendto(soc, pkt, plen, 0, so, so_sz);
 		if (e < 0)
-			perror("sendto ");
+			{
+			 perror("synscan->find_rtt(2)->sendto ");
+			 close(soc);
+			 bpf_close(bpf);
+			 return -1;
+			}
 		res = (char *) bpf_next_tv(bpf, &len, &tv);
 
 
@@ -489,7 +659,7 @@ find_rtt(struct in_addr dst, unsigned long *rtt)
 			if (noresend > 4)
 				noresend = 0;
 			if (err > 10) {
-				*rtt = htonl(2 << 28);
+				*rtt = htonl(1 << 28);
 				return 0;
 			}
 		}
@@ -497,59 +667,101 @@ find_rtt(struct in_addr dst, unsigned long *rtt)
 
 	close(soc);
 	bpf_close(bpf);
-        if(max == 0)max = htonl(2 << 28);
+        if(max == 0)max = htonl(1 << 28);
 	*rtt = max;
 	return 1;
 }
 
 
 struct list    *
-sendpacket(int soc, int bpf, int skip, struct in_addr dst, struct in_addr src, int dport, int magic, struct list * packets, struct timeval rtt_tv, int sniff, struct arglist * env)
+sendpacket(int soc, int bpf, int skip, struct in_addr dst, struct in_addr src, struct sockaddr_in6 * sin6, int dport, int magic, struct list * packets, unsigned long * rtt, int sniff, struct arglist * env)
 {
 	unsigned long   ack = maketime();
-	char           *pkt = mktcp(src, magic, dst, dport, ack);
+	int		plen;
+	char           *pkt = (is_ipv6 == 0) ? mktcp(src, magic, dst, dport, ack, 0, TH_SYN, &plen):mktcp6(sin6, magic, dport, ack, 0, TH_SYN, &plen);
 	int             len;
 	char           *res;
 	struct sockaddr_in soca;
+	struct sockaddr * so;
+	int    so_sz = 0;
+  	struct timeval rtt_tv = timeval(*rtt);
+	static int retry = 0;
 
 	bzero(&soca, sizeof(soca));
-	soca.sin_family = AF_INET;
-	soca.sin_addr = dst;
+
+	if ( is_ipv6 == 0 )
+	{
+	 soca.sin_family = AF_INET;
+	 soca.sin_addr = dst;
+	 so = (struct sockaddr*)&(soca);
+	 so_sz = sizeof(soca);
+	}
+	else
+	{
+	 so = (struct sockaddr*)sin6;
+	 so_sz = sizeof(*sin6);
+	}
+	rtt_tv.tv_sec *= 1000;
+	rtt_tv.tv_sec /= 8;
+	
+	rtt_tv.tv_usec += (rtt_tv.tv_sec % 1000) * 1000;
+	rtt_tv.tv_sec  /= 1000;
+        if ( rtt_tv.tv_sec >= 0 && rtt_tv.tv_usec > 10000 )
+		{
+		rtt_tv.tv_sec  = 0;
+		rtt_tv.tv_usec = 10000;
+		}
 
 	if (dport != 0) {
 		int             e;
+		ipci();
 		packets = add_packet(packets, dport, ack);
-		e = sendto(soc, pkt, sizeof(struct ip) + sizeof(struct tcphdr), 0, (struct sockaddr *) & soca, sizeof(soca));
+send:
+		errno = 0;
+		e = sendto(soc, pkt, plen, 0, so, so_sz);
 		if (e < 0) {
-			perror("sendto ");
-			close(soc);
-			bpf_close(bpf);
-			return NULL;
+			if ( errno == ENOBUFS )
+			{
+			 retry ++;
+			 if ( retry < 60 )
+				{
+				  sleep(1);
+				  goto send;
+				}
+			}
+			perror("synscan->sendpacket->sendto ");
+			return SCAN_FATAL_ERR;
 		}
 	}
 	if (sniff != 0) {
 		struct timeval  t, n;
 again:
-		/* gettimeofday(&t, NULL); */
 		res = (char *) bpf_next_tv(bpf, &len, &rtt_tv);
-		/*gettimeofday(&n, NULL);*/
-#if 0
-		if (n.tv_sec == t.tv_sec &&
-		    (n.tv_usec - t.tv_usec > rtt_tv.tv_usec))
-			printf("Woops - %ld (instead of %ld)\n", n.tv_usec - t.tv_usec, rtt_tv.tv_usec);
-#endif
 		if (res != NULL) {
 			unsigned short  sport = extractsport(res + skip, len);
 			int             synack = issynack(res + skip, len);
+			unsigned int rack = extractack(res + skip, len);
+			unsigned int rseq = extractseq(res + skip, len);
 			if (synack) {
+			  char * rst;
+			  int    len;
 #ifdef DEBUG
 				printf("=> Port %d is open\n", sport);
 #endif
-				scanner_add_port(env, sport, "tcp");
+		  	   scanner_add_port(env, sport, "tcp");
+			  /* Send a RST to make sure the connection is closed on the remote side */
+			  rst = (is_ipv6 == 0) ? mktcp(src, magic, dst, sport, htonl(ntohl(ack) + 1),htonl(ntohl(rseq) + 1 ), TH_RST, &plen):
+						 mktcp6(sin6, magic, sport, ack + 1, htonl(ntohl(rseq) + 1), TH_RST, &plen);
+			  sendto(soc, rst, sizeof(struct ip) + sizeof(struct tcphdr), 0, so, so_sz);
+			  /* Adjust the rtt */
+			  *rtt = compute_rtt(rack);
+			  if ( ntohl(*rtt) >= ( 1 << 28 ) ) *rtt = htonl(1 << 28);
+
 			}
+			ipcd();
 			packets = rm_packet(packets, sport);
 			rtt_tv.tv_sec = 0;
-			rtt_tv.tv_usec = 100;
+			rtt_tv.tv_usec = 0;
 			goto again;
 		}
 	}
@@ -557,7 +769,7 @@ again:
 }
 
 int 
-scan(struct arglist * env, struct in_addr dst, unsigned long rtt)
+scan(struct arglist * env, struct in_addr dst, struct sockaddr_in6 * sin6, unsigned long rtt)
 {
 	int             num;
 	int             soc = rawsocket();
@@ -566,7 +778,6 @@ scan(struct arglist * env, struct in_addr dst, unsigned long rtt)
 	int             magic = 4441 + (rand() % 1200);
 	int             skip;
 	int             i;
-	struct sockaddr_in soca;
 	struct list    *packets = NULL;
 	struct timeval  rtt_tv = timeval(htonl(ntohl(rtt) / 2));
 	struct arglist *globals = arg_get_value(env, "globals");
@@ -584,38 +795,54 @@ scan(struct arglist * env, struct in_addr dst, unsigned long rtt)
 
 	if (soc < 0)
 		return -1;
+#
+	if ( is_ipv6 != 0 )
+		bpf = openbpf6(sin6, magic);
+	else
+		bpf = openbpf(dst, &src, magic);
 
-	bpf = openbpf(dst, &src, magic);
 	skip = get_datalink_size(bpf_datalink(bpf));
-	bzero(&soca, sizeof(soca));
-	soca.sin_family = AF_INET;
-	soca.sin_addr = dst;
 
 
-	for (i = 0; ports[i]; i += 2) {
-		int             retry = 0;
+	for (i = 0; i < num ; i += 2) {
+		int  retry = 0;
 
+		
 
-		if (i % 100 == 0)
-			comm_send_status(globals, hname, "portscan", i, num);
-		packets = sendpacket(soc, bpf, skip, dst, src, ports[i], magic, packets, rtt_tv, 0, env);
-		packets = sendpacket(soc, bpf, skip, dst, src, ports[i + 1], magic, packets, rtt_tv, 1, env);
+   	  	if (i % 100 == 0)
+                        comm_send_status(globals, hname, "portscan", i, num);
+
+		if ( ipcc() != 0 ) 
+			packets = rm_dead_packets(packets, rtt, &retry);
+		else
+		{
+		int err = 0;
+		if ( sin6 == NULL && ! cct(dst, ports[i]) ) err = 1;
+		
+	         if ( err == 0 ) packets = sendpacket(soc, bpf, skip, dst, src, sin6, ports[i], magic, packets, &rtt, 0, env);
+		 if ( packets == SCAN_FATAL_ERR ) break;
+	 	 if ( i + 1 < num ) {
+		 	 packets = sendpacket(soc, bpf, skip, dst, src, sin6, ports[i + 1], magic, packets, &rtt, 1, env);
+			 if ( packets == SCAN_FATAL_ERR ) break;
+			}
+		}
 	}
 
 #ifdef DEBUG
 	printf("Done with the sending\n");
 #endif
 
-	while (packets != NULL) {
+	while (packets != NULL && packets != SCAN_FATAL_ERR ) {
 		i = 0;
 		retry = 0;
 		packets = rm_dead_packets(packets, rtt, &retry);
-		while (retry != 0 && i < 3) {
-			packets = sendpacket(soc, bpf, skip, dst, src, retry, magic, packets, rtt_tv, 0, env);
+		while (retry != 0 && i < 2) {
+			packets = sendpacket(soc, bpf, skip, dst, src, sin6, retry, magic, packets, &rtt, 0, env);
+			if ( packets == SCAN_FATAL_ERR ) break;
 			packets = rm_dead_packets(packets, rtt, &retry);
 			i++;
 		}
-		packets = sendpacket(soc, bpf, skip, dst, src, retry, magic, packets, rtt_tv, 1, env);
+		packets = sendpacket(soc, bpf, skip, dst, src, sin6, retry, magic, packets, &rtt, 1, env);
 	}
 
 	comm_send_status(globals, hname, "portscan", num, num);
@@ -650,7 +877,7 @@ PlugExport int
 plugin_init(struct arglist * desc)
 {
 	plug_set_id(desc, 11219);
-	plug_set_version(desc, "$Revision: 1.15 $");
+	plug_set_version(desc, "$Revision: 1.28 $");
 
 
 
@@ -674,14 +901,46 @@ plugin_run(struct arglist * env)
 	int             ret = maketime();
 	unsigned long   rtt;
 	struct in_addr *dst = plug_get_host_ip(env);
+	struct in_addr dest;
 	struct timeval  tv;
+	int kb_rtt;
+	int type;
+#ifdef IPV6_SUPPORT
+	struct sockaddr_in6 * sin6 = plug_get_host_ip6(env);
 
+	bzero(&dest, sizeof(dest));
+	if ( sin6 != NULL && dst == NULL ) is_ipv6 = 1;
+	else is_ipv6 = 0;
+	
+	if ( is_ipv6 != 0 && islocalhost6(&sin6->sin6_addr) ) return 0;
+#else
+	struct sockaddr_in * sin6 = NULL;
+#endif
 
+	if ( dst ) dest = *dst;
 
-	if (find_rtt(*dst, &rtt) < 0) {
-		fprintf(stderr, "Something went wrong, bailing out\n");
-		return (1);
+       
+       if ( (is_ipv6 == 0) && islocalhost(dst) ) return 0;
+
+       /* Don't scan for ever -- emergency fix : abort after one hour */
+       signal(SIGALRM, _exit);
+       alarm(3600);
+
+       kb_rtt = (int)plug_get_key ( env, "/tmp/ping/RTT", &type );
+       if ( kb_rtt )
+          rtt = (kb_rtt / 1000000) << 28 | ( ( (kb_rtt % 1000000 ) & 0xFFFFFFF0 ) >> 4 );
+       else
+	{
+	 if (find_rtt(dest, sin6, &rtt) < 0) 
+		{
+		 fprintf(stderr, "Something went wrong, bailing out\n");
+		 return (1);
+		}
 	}
+
+        if ( rtt >= ( 1 << 28 ) )
+	  rtt = htonl(1 << 28);
+
 #ifdef DEBUG
 	printf("RTT = 0x%.8x\n", ntohl(rtt));
 #endif
@@ -689,12 +948,15 @@ plugin_run(struct arglist * env)
 	tv = timeval(rtt);
 
 #ifdef DEBUG
-	printf("That %d seconds and %d usecs\n", tv.tv_sec, tv.tv_usec);
+	printf("That's %d seconds and %d usecs\n", tv.tv_sec, tv.tv_usec);
 #endif
 
 
-	scan(env, *dst, rtt);
+	scan(env, dest, sin6, rtt);
 	plug_set_key(env, "Host/scanned", ARG_INT, (void *) 1);
 	plug_set_key(env, "Host/scanners/synscan", ARG_INT, (void*)1);
 	return 0;
 }
+
+
+
